# MAPLE
## Features

- Performs MCMC based phylogenetics.
- Can handle polymorphism (or synonymous states of a concept).
- Can handle Jukes-Cantor, Felsenstein-81 models.
- Handles only Rooted trees like in linguistics. So, the trees will have branch lengths and a root.
- Handles trees like Phylo class in ape package of R.

## Parameter space moves
- Performs tree moves using external SPR and NNI.
- Branch lengths are sampled according to a Gamma-Dirichlet distribution.
- Substitution parameters are sampled using Dirichlet and DualSampler move.


## Usage
> Compile code using ```python3 setup.py build_ext --inplace```
```
 usage: san_scaled.py [-h] [-i INPUT_FILE] [-m MODEL] [-n N_GEN] [-t THIN]
                     [-d DATA_TYPE] [-o OUTPUT_FILE] [-N N_CHAINS]
                     [-ncats NR_CATEGORIES] [--ml_scale] [--ml_cache]
                     [-t0 INIT_TEMP]

optional arguments:
  -h, --help            show this help message and exit
  -i INPUT_FILE, --input_file INPUT_FILE
                        Input a file in Phylip format with taxa and characters
                        separated by a TAB character
  -m MODEL, --model MODEL
                        JC/F81/GTR
  -n N_GEN, --n_gen N_GEN
                        Number of generations
  -t THIN, --thin THIN  Number of generations after to print to screen
  -d DATA_TYPE, --data_type DATA_TYPE
                        Type of data if it is binary/multistate. Specify bin
                        for binary and multi for multistate characters or
                        phonetic alignments
  -o OUTPUT_FILE, --output_file OUTPUT_FILE
                        Name of the out file prefix
  -N N_CHAINS, --n_chains N_CHAINS
                        Name of the out file prefix
  -ncats NR_CATEGORIES, --nr_categories NR_CATEGORIES
                        Number of categories for discrete Gamma distribution
  --ml_scale
  --ml_cache
  -t0 INIT_TEMP, --init_temp INIT_TEMP
                        Initial Temperature
  
  ```
  Linguistic datasets are quite different from biological morphological datasets. A Jukes-Cantor model where all the transition rates are the same is useful for estimating trees with branch lengths.

Example Usage:
```
#python3 san.py -i fcd_glotto_phy/ccm--aa-58-200.phy -m F81 -t 20 -d bin -o ccm-aa-san_2.10 -t0 50 > ccm-aa-san.2.10.log.txt
python3 san_scaled.py -i fcd_glotto_phy/ccm--aa-58-200.phy -m F81 -t 40 -d bin -o ccm-aa-san_40.80 -t0 100 --ml_scale --ml_cache
```  

# Requirements
- Python3, Cython, numpy, scipy 
