#from __future__ import print_function
import argparse, utils, time
from mcmc_gamma import *
#from ML_gamma import *
from ML_scaled_cache import *
import config, sys
from collections import defaultdict
np.random.seed(1234)
random.seed(1234)

T_i, mult_scale, T_stop = 20, 0.95, 10e-5 # Log scheduling does not work so well

time_start = time.time()

parser = argparse.ArgumentParser()
parser.add_argument("-i", "--input_file", help="Input a file in Phylip format with taxa and characters separated by a TAB character",  type=str)
parser.add_argument("-m", "--model", help="JC/F81/GTR",  type=str, default = "F81")
parser.add_argument("-n","--n_gen", help="Number of generations",  type=int, default = 100000)
parser.add_argument("-t","--thin", help="Number of generations after to print to screen",  type=int, default = 500)
parser.add_argument("-d","--data_type", help="Type of data if it is binary/multistate. Specify bin for binary and multi for multistate characters or phonetic alignments",  type=str)
parser.add_argument("-o","--output_file", help="Name of the out file prefix",  type=str, default = "temp")
parser.add_argument("-N","--n_chains", help="Name of the out file prefix",  type=int, default=4)
#parser.add_argument("-dt","--deltaT", help="Chain temperature spacing",  type=float, default=0.1)
#parser.add_argument("-a","--adjust_step", help="Chain temperature spacing",  type=int, default=500)
#parser.add_argument("-C","--st_const", help="Upper bound",  type=int, default=10)
#parser.add_argument("-p","--inc_temp", help="Harmonic/geometric spacing of temperature",  type=str, default="hm")
#parser.add_argument("-cr","--converge_ratio", help="Convergence ratio",  type=float, default=1.5)
#parser.add_argument("-ct","--cold_chain_thin", help="Thin the cold chain",  type=int, default=500)
#parser.add_argument("-cns","--cold_n_samples", help="Number of samples in cold chain",  type=int, default = 1000)
parser.add_argument("-ncats","--nr_categories", help="Number of categories for discrete Gamma distribution",  type=int, default = 4)
parser.add_argument('--ml_scale', default=False, action='store_true')
parser.add_argument('--ml_cache', default=False, action='store_true')
parser.add_argument("-t0","--init_temp", help="Initial Temperature",  type=int)
#parser.add_argument("-sd","--seed", help="Initial Seed Value",  type=int, default = 1)
args = parser.parse_args()


if args.n_chains <= 1:
    print("Number of chains should be greater than 1")
    sys.exit(1)

if args.data_type == "bin":
    config.N_TAXA, config.N_CHARS, config.ALPHABET, site_dict, config.LEAF_LLMAT, config.TAXA, config.N_SITES = utils.readBinaryPhy(args.input_file)

elif args.data_type == "multi":
    config.N_TAXA, config.N_CHARS, config.ALPHABET, site_dict, config.LEAF_LLMAT, config.TAXA, config.N_SITES = utils.readMultiPhy(args.input_file)


config.IN_DTYPE = args.data_type
config.N_GEN = args.n_gen
config.THIN = args.thin
config.MODEL = args.model
config.N_NODES = 2*config.N_TAXA -1
config.N_CATS = args.nr_categories
T_i = args.init_temp

print("Characters ", config.N_CHARS)
print("Sites ", config.N_SITES)
print("Number of TAXA ", config.N_TAXA)
print("Alphabet ", config.ALPHABET)
print("Number of Discrete Gamma Categories ", config.N_CATS)

n_rates = config.N_CHARS*(config.N_CHARS-1)//2
prior_pi = np.array([1]*config.N_CHARS)
prior_er = np.array([1]*n_rates)

if config.MODEL == "JC":
    config.NORM_BETA = config.N_CHARS/(config.N_CHARS-1)

init_state = state_init()
site_rates = get_siterates(init_state["srates"])

print("Frequencies ", np.array(init_state["pi"]))
print("Rates ", np.array(init_state["rates"]))
print("Site rates ", np.array(site_rates))
print("Discrete Gamma distn shape parameter", init_state["srates"])
print("Root of the tree", init_state["root"])

if args.ml_scale:
    print("Scaled Likelihood")
    init_state["logLikehood"], cache_LL_Mats, cache_rescale_list = matML_scaled(init_state["pi"], init_state["root"], config.LEAF_LLMAT, init_state["postorder"], init_state["transitionMat"], config.N_SITES, config.N_TAXA, config.N_CATS)
else:
    print("Not scaled Likelihood")
    init_state["logLikehood"], _ = matML(init_state["pi"], init_state["root"], config.LEAF_LLMAT, init_state["postorder"], init_state["transitionMat"], config.N_SITES, config.N_TAXA, config.N_CATS)

state = init_state.copy() #Why am I copying?
initial_tree = adjlist2newickBL(state["tree"], adjlist2nodes_dict(state["tree"]), state["root"])+";"
cache_paths_dict = adjlist2reverse_nodes_dict(state["tree"]) #for nodes recompute

#print("Initial Random Tree ", initial_tree, sep="\t") #Comment for small number of taxa

print("Initial Likelihood ",state["logLikehood"])

#Set the weights for MCMC moves
if config.MODEL == "F81":
    params_list = ["pi", "tree", "bl", "srates"]
    weights = np.array([0.5, 5, 1, 0.5], dtype=np.float64)
elif config.MODEL == "GTR":
    params_list = ["pi","rates", "tree", "bl", "srates"]
    weights = np.array([0.5, 1.0, 3, 4, 0.5], dtype=np.float64)
    if state["pi"].shape[0] == 2: #Hack the rates move in GTR model.
        weights[1] = 0.0
    print("Weights ", weights)
elif config.MODEL == "JC":
    params_list = ["tree", "bl", "srates"]
    weights = np.array([3, 4, 0.5], dtype=np.float64)

tree_move_weights = np.array([4, 1], dtype=np.float64)
bl_move_weights = np.array([3, 0, 1], dtype=np.float64)

weights = weights/np.sum(weights)
tree_move_weights = tree_move_weights/np.sum(tree_move_weights)
bl_move_weights = bl_move_weights/np.sum(bl_move_weights)

moves_count = defaultdict(int)
accepts_count = defaultdict(int)
moves_dict = {"pi": [mvDualSlider], "rates": [mvRatesSlider], "tree":[rooted_NNI, externalSPR], "bl":[scale_edge, node_slider, slide_edge], "srates":[mvShapeScaler]}

if args.data_type == "bin":
    moves_dict["pi"] = [mvBinaryDualSlider]

params_fileWriter = open(args.output_file+".log","w")
trees_fileWriter = open(args.output_file+".trees","w")
const_states = ["pi("+idx+")" for idx in config.ALPHABET]

propose_state, prop_tmats = {}, []


print("Iter", "LnL", "TL", "Alpha", "Temperature", sep="\t", file=params_fileWriter)

n_iter = 1

while(1):

    propose_state["pi"],  propose_state["rates"], propose_state["srates"], propose_state["tree"], propose_state["postorder"] = state["pi"].copy(), state["rates"].copy(), state["srates"], state["tree"].copy(), state["postorder"].copy()
    
    current_ll, proposed_ll, ll_ratio, hastings_ratio, change_edge, pr_ratio, old_edge_p_ts, old_edge_p_t_as = 0.0, 0.0, 0.0, 0.0, None, 0.0, [], []
        
    param_select = np.random.choice(params_list, p=weights)

    if param_select == "tree":
        move = np.random.choice(moves_dict[param_select], p=tree_move_weights)
    elif param_select == "bl":
        move = np.random.choice(moves_dict[param_select], p=bl_move_weights)
    else:
        move = np.random.choice(moves_dict[param_select])
    
    moves_count[param_select,move.__name__] += 1
    
    if param_select in ["pi", "rates"]:
        propose_param, hastings_ratio = move(state[param_select].copy())
        propose_state[param_select] = propose_param

    elif param_select == "bl":

        if move.__name__ == "scale_edge" or move.__name__ == "slide_edge":
            prop_edges_dict, hastings_ratio, pr_ratio, change_edge = move(state["tree"].copy())

        elif move.__name__ == "node_slider":
            prop_edges_dict, hastings_ratio, pr_ratio, change_edge, change_parent_edge = move(state["tree"].copy(), state["root"])

        propose_state["tree"] = prop_edges_dict

        nodes_recompute = get_path2root(cache_paths_dict, change_edge[1], state["root"]) #Recompute only for these nodes

    elif param_select == "tree":

        if move.__name__ == "rooted_NNI":
            prop_edges_dict, prop_post_order, hastings_ratio, nodes_recompute, nodes_list = move(state[param_select].copy(), state["root"])
        else:
            prop_edges_dict, prop_post_order, hastings_ratio = move(state[param_select].copy(), state["root"])

        propose_state["tree"] = prop_edges_dict
        propose_state["postorder"] = prop_post_order

    elif param_select == "srates":
        propose_param, hastings_ratio, pr_ratio = move(state[param_select])
        propose_state[param_select] = propose_param
        current_rates = site_rates[:]        
        site_rates = get_siterates(propose_param)

    #cache_rescale_list is refreshed when performing a tree move or substitution model parameters moves

    if move.__name__ == "scale_edge" or move.__name__ == "slide_edge":
        for imr, mean_rate in enumerate(site_rates):
            old_edge_p_ts.append(state["transitionMat"][imr][change_edge].copy())
            state["transitionMat"][imr][change_edge] = get_edge_transition_mat(propose_state["pi"], propose_state["rates"], propose_state["tree"][change_edge]*mean_rate)

        if args.ml_scale:
            if not args.ml_cache:
                proposed_ll, proposed_llMat, prop_rescale_list = matML_scaled(propose_state["pi"],  state["root"], config.LEAF_LLMAT, state["postorder"], state["transitionMat"], config.N_SITES, config.N_TAXA, config.N_CATS)
            else:
                proposed_ll, proposed_llMat, prop_rescale_list = matML_scaled_cache(propose_state["pi"],  state["root"], config.LEAF_LLMAT, state["postorder"], state["transitionMat"], config.N_SITES, config.N_TAXA, config.N_CATS, cache_LL_Mats, cache_rescale_list, nodes_recompute)
            
        else:
            proposed_ll, _ = matML(propose_state["pi"],  state["root"], config.LEAF_LLMAT, state["postorder"], state["transitionMat"], config.N_SITES, config.N_TAXA, config.N_CATS)

    elif move.__name__ == "node_slider":
        for imr, mean_rate in enumerate(site_rates):
            old_edge_p_ts.append(state["transitionMat"][imr][change_edge].copy())
            old_edge_p_t_as.append(state["transitionMat"][imr][change_parent_edge].copy())

            state["transitionMat"][imr][change_edge] = get_edge_transition_mat(propose_state["pi"], propose_state["rates"], propose_state["tree"][change_edge]*mean_rate)
            state["transitionMat"][imr][change_parent_edge] = get_edge_transition_mat(propose_state["pi"], propose_state["rates"], propose_state["tree"][change_parent_edge]*mean_rate)

        if args.ml_scale:
#                proposed_ll, proposed_llMat, prop_rescale_list = matML_scaled(propose_state["pi"],  state["root"], config.LEAF_LLMAT, state["postorder"], state["transitionMat"], config.N_SITES, config.N_TAXA, config.N_CATS)
            if not args.ml_cache:
                proposed_ll, proposed_llMat, prop_rescale_list = matML_scaled(propose_state["pi"],  state["root"], config.LEAF_LLMAT, state["postorder"], state["transitionMat"], config.N_SITES, config.N_TAXA, config.N_CATS)
            else:
                proposed_ll, proposed_llMat, prop_rescale_list = matML_scaled_cache(propose_state["pi"],  state["root"], config.LEAF_LLMAT, state["postorder"], state["transitionMat"], config.N_SITES, config.N_TAXA, config.N_CATS, cache_LL_Mats, cache_rescale_list, nodes_recompute)
        else:
            proposed_ll, _ = matML(propose_state["pi"],  state["root"], config.LEAF_LLMAT, state["postorder"], state["transitionMat"], config.N_SITES, config.N_TAXA, config.N_CATS)

    else:
        prop_tmats = [get_prob_t(propose_state["pi"], propose_state["tree"], propose_state["rates"], mean_rate) for mean_rate in site_rates]

        if args.ml_scale:
            proposed_ll, proposed_llMat, prop_rescale_list = matML_scaled(propose_state["pi"],  state["root"], config.LEAF_LLMAT, propose_state["postorder"], prop_tmats, config.N_SITES, config.N_TAXA, config.N_CATS)
        else:
            proposed_ll, _ = matML(propose_state["pi"],  state["root"], config.LEAF_LLMAT, propose_state["postorder"], prop_tmats, config.N_SITES, config.N_TAXA, config.N_CATS)


    current_ll = state["logLikehood"]
    ll_ratio = proposed_ll - current_ll + pr_ratio
    ll_ratio = (1.0/T_i)*ll_ratio
    ll_ratio += hastings_ratio

    if np.log(np.random.random()) <= ll_ratio:
        if param_select == "bl":
            state["tree"] = prop_edges_dict#propose_state["tree"]
        elif param_select == "tree":
            state[param_select] = prop_edges_dict#propose_state[param_select]
            state["postorder"] = prop_post_order
            cache_paths_dict = adjlist2reverse_nodes_dict(state[param_select])
        else:
            state[param_select] = propose_param#propose_state[param_select]
#                if param_select == "srates":
#                    print(param_select, propose_param)

        if move.__name__ not in ["scale_edge", "node_slider", "slide_edge"]:
            state["transitionMat"] = prop_tmats

        state["logLikehood"] = proposed_ll

        accepts_count[param_select,move.__name__] += 1

        cache_LL_Mats = proposed_llMat
        cache_rescale_list = prop_rescale_list

    else:
        if param_select == "srates":
            site_rates = current_rates[:]
        elif move.__name__ == "scale_edge" or move.__name__ == "slide_edge":
            for ip_t, p_t in enumerate(state["transitionMat"]):
                state["transitionMat"][ip_t][change_edge] = old_edge_p_ts[ip_t]
        elif move.__name__ == "node_slider":
            for ip_t, p_t in enumerate(state["transitionMat"]):
                state["transitionMat"][ip_t][change_edge] = old_edge_p_ts[ip_t]
                state["transitionMat"][ip_t][change_parent_edge] = old_edge_p_t_as[ip_t]


    if n_iter % config.THIN == 0:
        TL = sum(state["tree"].values())
        stationary_freqs = "\t".join([str(state["pi"][idx]) for idx in range(config.N_CHARS)])
        sampled_tree = adjlist2newickBL(state["tree"], adjlist2nodes_dict(state["tree"]), state["root"])+";"
        print(n_iter, current_ll, proposed_ll, TL, param_select, move.__name__, T_i, sep="\t")
        print(n_iter, state["logLikehood"], TL, state["srates"], sep="\t", file=params_fileWriter)
        print(n_iter, sampled_tree, sep="\t", file=trees_fileWriter)
        if T_i < T_stop: break
        T_i = T_i * mult_scale 


    n_iter +=  1

for k, v in moves_count.items():
    print(*k, round(accepts_count[k]/v, 2))

time_end = time.time()

print("MCMC chain run for {} iterations for {} seconds".format(n_iter, round(time_end-time_start, 3)))


