# Data and source code, accompanying the paper "An automatic framework for fast cognate detection and Bayesian phylogenetic inference in computational historical linguistics"

When using this software, please cite the original paper where we introduce it:

> Rama, T. and List, J.-M. (2019): An automatic framework for fast cognate detection and Bayesian phylogenetic inference in computational historical linguistics. Proceedings of the 57th Annual Meeting of the Association for Computational Linguistics.

# 1 BipSkip: An algorithm for fast cognate detection

The Python package to run the algorithm is provided in the folder `1_BipSkip` of this repository. In order to install it, you will need to install the `igraph` package independently, including the Python version. To install BipSkip with dependencies itself, you can use `pip`, when opening a terminal in the folder:

```
$ pip install -r pip-requirements.txt
```

Alternatively, you can install the requirements automatically:

```
$ python setup.py develop
```

The algorithm is designed as a LingPy pluging. You can see how it is used by inspecting the code described in the following.

# 2 Training the BipSkip algorithm

As of now, BipSkip has a couple of parameters which may have a huge impact on
the results. In order to train these, we conducted an exhaustive search on all possible parameter combinations, using the training-data
by @List2017c, originally published by @List2014d. 

To replicate this training analysis, just open a terminal in the folder `2_Training`, and type:

```
$ python training.py
```

# 3 Testing BipSkip

To test the BipSkip algorithm in the different variants, just type the following in the terminal:

```
$ python test.py test # normal analysis, test data by List et al. 2017
$ python test.py test2 # normal analysis, ttest data by Rama et al. 2018
$ python test.py test # infomap analysis, test data by List et al. 2017
$ python test.py test2 # infomap analysis, test data by Rama et al. 2018
$ python test.py sca # sca analysis, test data by List et al. 2017
$ python test.py sca2 # sca analysis, test data by Rama et al. 2018
$ python test.py ccm # ccm analysis, test data by List et al. 2017
$ python test.py ccm2 # ccm analysis, test data by Rama et al. 2018

```

In all cases, by adding `nex` as a keyword, you can further export your data to nexus format for the further usage in the MAPLE workflow.

# 4 Testing MAPLE

Detailed instructions on how to use MAPLE can be found in the file `4_MAPLE/src/README.md`.

# 5 Results

We list the results in a spreadsheet file. Detailed results, however, can be found also in the text files in the four folders where we applied our experiments.
